'''
Version and license information.
'''

__all__ = ['__version__', '__versiondate__', '__license__']

__version__ = '0.3.3'
__versiondate__ = '2024-04-16'
__license__ = f'Starsim {__version__} ({__versiondate__}) — © 2023-2024 by IDM'
